import logging
import socket
from datetime import timedelta
from time import sleep
from typing import Union, List

from django.db import transaction
from django.utils import timezone
from privex.helpers import empty, PrivexException
from lockmgr.models import Lock

log = logging.getLogger(__name__)


class Locked(PrivexException):
    """Raised when a lock already exists with the given name"""
    pass


class LockMgr:
    """

    Usage:

        Using a ``with`` statement, create a LockMgr for ``mylock`` with automatic expiration if held for more than
        60 seconds. After the ``with`` statement is completed, all locks created will be removed.

        >>> try:
        ...     with LockMgr('mylock', 60) as l:
        ...         print('Doing stuff with mylock locked.')
        ...         # Obtain an additional lock for 'otherlock' - will use the same expiry as mylock
        ...         # Since ``ret`` is set to True, it will return a bool instead of raising Lock
        ...         if l.lock('otherlock', ret=True):
        ...             print('Now otherlock is locked...')
        ...             l.unlock('otherlock')
        ...         else:
        ...             print('Not doing stuff because otherlock is already locked...')
        ... except Locked as e:
        ...     print('Failed to lock. Reason: ', type(e), str(e))

    """

    def __init__(self, name, expires: int = 600, locked_by=None, lock_process=None, wait: int = None):
        self.expires = int(expires)
        self.wait = None if empty(wait) else int(wait)
        self.locked_by = socket.gethostname() if empty(locked_by) else locked_by
        self.name, self.lock_process = name, lock_process
        self._locks = []  # type: List[Lock]

    def lock(self, name, expires: int = None, ret: bool = False, wait: int = None):
        """
        Obtains a lock using :py:func:`.get_lock` and appends it to :py:attr:`._locks` if successful.

        If the argument ``ret`` is ``False`` (default), it will raise :class:`.Locked` if the lock couldn't be obtained.

        Otherwise, if ``ret`` is ``True``, it will simply return ``False`` if the requested lock name is already locked.

        :param str name: A unique name to identify your lock
        :param int expires: (Default: 600 sec) How long before this lock is considered stale and forcefully released?
        :param bool ret: (Default: False) Return ``False`` if locked, instead of raising ``Locked``.
        :param int wait: (Optional) Retry obtaining the lock for this many seconds. MUST be divisible by 5.
                         If not empty, will retry obtaining the lock every 5 seconds until ``wait`` seconds

        :raises Locked: If the requested lock ``name`` is already locked elsewhere, :class:`.Locked` will be raised

        :return bool success: ``True`` if successful. If ``ret`` is true then will also return False on failure.
        """
        expires = self.expires if empty(expires) else int(expires)
        wait = self.wait if empty(wait) else int(wait)
        log.debug('Attempting to get lock %s with expiry of %s', name, expires)
        try:
            lck = get_lock(name=name, locked_by=self.locked_by, lock_process=self.lock_process, expires=expires)
            self._locks.append(lck)
            log.debug('Lock obtained for %s', name)
            return True
        except Locked as e:
            if empty(wait, zero=True):
                log.info('A lock already exists on %s...', name)
                if ret: return False
                raise e

            wait = int(wait)
            if wait % 5 != 0:
                raise ArithmeticError('The argument "wait" must be divisible by 5 seconds.')
            log.info('Lock "%s" was locked - waiting up to %s seconds for lock to be released.', name, wait)
            log.info('Retrying lock in 5 seconds...')
            sleep(5)
            return self.lock(name=name, expires=expires, ret=ret, wait=wait - 5)

    @staticmethod
    def unlock(lock: Union[Lock, str]):
        """Alias for :py:func:`.unlock`"""
        return unlock(lock)

    def __enter__(self):
        self.lock(self.name, self.expires)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        log.debug('LockMgr exiting. Releasing all held locks')
        for lock in self._locks:
            self.unlock(lock)


def is_locked(name) -> bool:
    """Cleans expired locks, then returns ``True`` if the given lock key ``name`` exists, otherwise ``False``"""
    clean_locks()
    with transaction.atomic():
        lck = Lock.objects.select_for_update().filter(name=name)
    return lck.count() > 0


def clean_locks():
    """Deletes expired :class:`payments.models.Lock` objects."""
    with transaction.atomic():
        expired_locks = Lock.objects.select_for_update().filter(locked_until__lt=timezone.now())
        if expired_locks.count() > 0:
            log.info('Deleting expired locks: %s', expired_locks.values_list('name', flat=True))
            expired_locks.delete()


def get_lock(name, expires: int = 600, locked_by: str = None, lock_process: int = None) -> Lock:
    """
    READ THIS: It's best to use :class:`.LockMgr` as it automatically handles locking and unlocking using ``with``.

    Calls :py:func:`.clean_locks` to remove any expired locks, checks for any existing locks using a FOR UPDATE
    transaction, then attempts to obtain a lock using the Lock model :class:`payments.models.Lock`

    If ``name`` is already locked, then :class:`.Locked` will be raised.

    Otherwise, if it was successfully locked, a :class:`payments.models.Lock` object for the requested lock name
    will be returned.

    Usage:

        >>> try:   # Obtain a lock on 'mylock', with an automatic expiry of 60 seconds.
        ...     mylock = get_lock('mylock', 60)
        ...     print('Successfully locked mylock')
        ... except Locked as e:
        ...     print('Failed to lock. Reason: ', type(e), str(e))
        ... finally:  # Regardless of whether there was an exception or not, remember to remove the lock!
        ...     print('Removing lock on "mylock"')
        ...     unlock(mylock)

    :param str name: A unique name to identify your lock
    :param int expires: (Default: 600 sec) How long before this lock is considered stale and forcefully released?
    :param str locked_by: (Default: system hostname) What server/app is trying to obtain this lock?
    :param int lock_process: (Optional) The process ID requesting the lock

    :raises Locked: If the requested lock ``name`` is already locked elsewhere, :class:`.Locked` will be raised

    :return Lock lock: If successfully locked, will return the :class:`payments.models.Lock` of the requested lock.

    """
    # First let's remove any old expired locks
    clean_locks()
    locked_by = socket.gethostname() if empty(locked_by) else locked_by
    is_locked = Lock.objects.filter(name=name)
    if is_locked.count() > 0:
        raise Locked(f'Lock with name {name} already exists.')
    with transaction.atomic():
        is_locked = Lock.objects.select_for_update().filter(name=name)
        if is_locked.count() > 0:
            raise Locked(f'Lock with name {name} already exists.')
        lock = Lock.objects.create(
            name=name, locked_by=locked_by, lock_process=lock_process,
            locked_until=timezone.now() + timedelta(seconds=int(expires))
        )
        lock.save()
        return lock
    raise Locked(f'Unexpected transaction.atomic exit via get_lock()')


def unlock(lock: Union[Lock, str]):
    """
    Releases a given lock - either specified as a string name, or as a :class:`payments.models.Lock` object.

    Usage:

        >>> mylock = get_lock('mylock', expires=60)
        >>> unlock('mylock') # Delete the lock by name
        >>> unlock(mylock)   # Or by Lock object.

    :param lock:
    :return:
    """
    log.debug('Releasing lock for %s', lock)
    if type(lock) is str:
        Lock.objects.filter(name=lock).delete()
        return True
    lock.delete()
    return True



from privex.loghelper import LogHelper
import os
from os import getenv as env
import logging

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
DEBUG = False
SECRET_KEY = 'NotApplicable'

LOG_FORMATTER = logging.Formatter('[%(asctime)s]: %(name)-55s -> %(funcName)-20s : %(levelname)-8s:: %(message)s')

LOG_LEVEL = logging.WARNING

_lh = LogHelper('lockmgr', formatter=LOG_FORMATTER, handler_level=LOG_LEVEL)

_lh.add_console_handler()

INSTALLED_APPS = ['django_nose', 'lockmgr']

DATABASES = {}

TEST_RUNNER = 'django_nose.NoseTestSuiteRunner'

if env('DB_BACKEND', 'sqlite') in ['sqlite', 'sqlite3']:
    DATABASES = dict(default=dict(
        ENGINE='django.db.backends.sqlite3',
        NAME=os.path.join(BASE_DIR, env('DB_PATH', 'db.sqlite3'))
    ))
else:
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.' + env('DB_BACKEND'),
            'NAME': env('DB_NAME', 'lockmgr'),
            'USER': env('DB_USER', 'lockmgr'),
            'PASSWORD': env('DB_PASS', ''),
            'HOST': env('DB_HOST', 'localhost'),
        }
    }



# Generated by Django 2.2.4 on 2019-09-01 11:30

from django.db import migrations, models
import lockmgr.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Lock',
            fields=[
                ('name', models.CharField(max_length=100, primary_key=True, serialize=False, verbose_name='Unique Lock Name')),
                ('locked_by', models.CharField(blank=True, max_length=100, null=True)),
                ('lock_process', models.IntegerField(blank=True, default=0, null=True, verbose_name='Process ID of lock requester')),
                ('locked_until', models.DateTimeField(default=lockmgr.models.default_lock_expiry)),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Locked At')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Last Update')),
            ],
        ),
    ]



from django.db import models
from django.utils import timezone
from datetime import timedelta


def default_lock_expiry():
    return timezone.now() + timedelta(minutes=60)


class Lock(models.Model):
    name = models.CharField('Unique Lock Name', max_length=100, primary_key=True)
    """Unique name of the lock, referring to what specific resource(s) is locked"""

    locked_by = models.CharField(max_length=100, blank=True, null=True)
    """Name of the node / app which created this lock"""

    lock_process = models.IntegerField('Process ID of lock requester', default=0, blank=True, null=True)

    locked_until = models.DateTimeField(default=default_lock_expiry)
    """
    Locks have an expiration time, to help avoid the issue of stuck locks, either due to forgetting to add cleanup code, 
    or simply due to the app/server crashing before it can release the lock.

    After a lock has expired, it's assumed that the lock is stale and needs to be removed, and the affected resources 
    are safe to use.
    """

    created_at = models.DateTimeField('Locked At', auto_now_add=True)
    updated_at = models.DateTimeField('Last Update', auto_now=True)

    def __str__(self):
        return f"<Lock name='{self.name}' locked_by='{self.locked_by}' locked_until='{self.locked_until}'>"



"""
+===================================================+
|                 © 2019 Privex Inc.                |
|               https://www.privex.io               |
+===================================================+
|                                                   |
|        Django Database Lock Manager               |
|        License: X11/MIT                           |
|                                                   |
|        Core Developer(s):                         |
|                                                   |
|          (+)  Chris (@someguy123) [Privex]        |
|                                                   |
+===================================================+

Django Database Lock Manager - Easy to use lock system using your Django app's database
Copyright (c) 2019    Privex Inc. ( https://www.privex.io )

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of
the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name(s) of the above copyright holders shall not be used in advertising or
otherwise to promote the sale, use or other dealings in this Software without prior written authorization.

"""
# from django.core.exceptions import AppRegistryNotReady
#
# try:
#     from lockmgr.models import Lock
#     from lockmgr.lockmgr import LockMgr, Locked, get_lock, unlock
# except AppRegistryNotReady:
#     pass

name = 'lockmgr'

VERSION = '1.1.2'

